import numpy as np
import pandas as pd
#figure factory for tables
import plotly.figure_factory as ff
from plotly.offline import download_plotlyjs, init_notebook_mode, plot, iplot
from plotly.graph_objs import Scatter, Figure, Layout
import plotly.graph_objs as go
import plotly
init_notebook_mode(connected=False)

import cufflinks as cf
cf.set_config_file(offline=True, world_readable=False,offline_show_link=False,theme='pearl')

def tsplot(obj,title='',xTitle='Date',yTitle='',asFigure=False):
    """Plots actual, fitted, and predicted values from forecast class in plotly graph
    Args:
        obj: forecast object (dict for single series and dataframe for multiple series)
        title: plot title
        xTitle: xaxis title
        yTitle: y axis title
        asFigure: parameter to return as fig instead of the plot if set to True
        
    Returns:
        fig: plotly figure object
        
    
    """
    
    try:
        #handle dict (assumes it's forecast object)
        if isinstance(obj, dict):
            trace_actuals = go.Scatter(
                x = obj['x'].index,
                y = obj['x'].values,
                mode = 'lines',
                name = 'actual'
            )
            trace_fitted = go.Scatter(
                x = obj['fitted'].index,
                y = obj['fitted'].values,
                mode = 'lines+markers',
                name = 'fitted',
                opacity=0.8
            )

            trace_predicted = go.Scatter(
                x = obj['predicted'].index,
                y = obj['predicted'].values,
                mode='lines',
                name = 'predicted'
            )

            trace_lower = go.Scatter(
                x=obj['predicted'].index,
                y=obj['lower'],
                fill= None,
                mode='lines',
                name='Lower PI ('+str(obj['level'])+'%)',
                line=dict(
                    color='lightgreen',
                )
            )
            trace_upper = go.Scatter(
                x=obj['predicted'].index,
                y=obj['upper'],
                fill='tonexty',
                mode='lines',
                name='Upper PI ('+str(obj['level'])+'%)',
                line=dict(
                    color='lightgreen',
                )
            )

            data = [trace_actuals, trace_fitted,trace_predicted,trace_lower,trace_upper]

            layout = dict(
                    title=title,
                    yaxis = dict(title = yTitle),
                    xaxis=dict(
                        title = xTitle,
                        rangeslider=dict(),
                        type='date'
                    )
                )
            fig = dict(data=data, layout=layout)

            
        elif isinstance(obj, pd.core.series.Series):
            
            trace_series = go.Scatter(
                x = obj.index,
                y = obj.values,
                mode = 'lines',
                name = 'series'
            )
            data = [trace_actuals]
            
            layout = dict(
                    title=title,
                    yaxis = dict(title = yTitle),
                    xaxis=dict(
                        title = xTitle,
                        rangeslider=dict(),
                        type='date'
                    )
            )
            fig = dict(data=data, layout=layout)

            
        #separate logic if input is dataframe
        elif isinstance(obj, pd.core.frame.DataFrame):
            #return cufflinks figure and add rangeslider to layout
            fig = obj.iplot(kind='scatter',title=title,yTitle=yTitle,xTitle=xTitle,asFigure=True) 
            fig['layout']['xaxis1']['rangeslider'] = dict()

        else:
            raise TypeError('only accepted objects are dictionary or dataframe returned from forecast class')
    except TypeError:
        print('only accepted objects are dictionary or dataframe returned from forecast class')
        
    if asFigure == True:
        return fig
    return iplot(fig, show_link=False)
    
def acc_plot(obj,title='',xTitle='Date',yTitle='',mode='lines+markers',tablewidth=350,asFigure=False):
    
    """ plots accuracy measures
    Args:
        obj: accuracy object(dict for single series, dataframe for multiple series)
        title: plot title
        xTitle: xaxis title
        yTitle: y axis title
        mode: one of 'lines', 'markers', or 'lines+markers' (only applies to multiple series)
        tablewidth:sets table width for accuracy measures of single instance
        asFigure: parameter to return as fig instead of the plot if set to True
        
    Returns:
        fig: plotly figure object
        
    
    """
    
    try:
        #handle dict (assumes it's forecast object)
        if isinstance(obj, dict):
            #create dataframe from accuracy dictionary
            table_df = pd.DataFrame.from_dict(obj, orient='index', dtype=None)
            table_df.columns = ['Value']
            #round decimals
            table_df = table_df.round(5)
            fig = ff.create_table(table_df, index=True, index_title='Accuracy Measure')
            fig.layout.update({'width':tablewidth})

            
        #separate logic if input is dataframe
        elif isinstance(obj, pd.core.frame.DataFrame):
            #return cufflinks figure, normalize error measures, then plot them without a y axis
            normalized_df=obj.T.apply(lambda x: x/x.max(), axis=0)
            fig = normalized_df.iplot(kind='scatter',mode=mode,asFigure=True)
            fig['layout']['yaxis1']['visible'] = False
            for error_measure in fig['data']:
                #get correct text values from original dataframe and round values to 4 decimals
                error_measure['text'] = np.around(obj.T[error_measure['name']].values,decimals=4)
                error_measure['hoverinfo'] = 'text+name'

        else:
            raise TypeError('only accepted objects are dictionary or dataframe returned from forecast class')
    except TypeError:
        print('only accepted objects are dictionary or dataframe returned from forecast class')
        
    if asFigure == True:
        return fig
    return iplot(fig, show_link=False)



import numpy as np
import pandas as pd
import dask

def accuracy(actual,predicted=None,separate_series=False):
    
    """returns accuracy measures
    Args:
        actual: actual values as numpy array or time series
        predicted: predicted values as numpy array or time series
        separate_series: only used for data frames. if set to true, will calculate accuracy metrics 
            for each series in parallel using dask and return as dataframe instead of overall accuracy
        
    Returns:
        accuracy_dict: dictionary of accuracy measures
        
    
    prints accuracy measures for forecasting and returns them as a dictionary
    """
    #handle different incoming data types
    try:
        #handle dict (assumes it's forecast object)
        if isinstance(actual, dict):
            predicted = actual['fitted'].values.astype(float)
            actual = actual['x'].values.astype(float)
            actual[np.isnan(actual)] = 0.0
            predicted[np.isnan(predicted)] = 0.0
            
        elif isinstance(predicted, np.ndarray):
            predicted = predicted.astype(float)
            predicted[np.isnan(predicted)] = 0.0
            
        elif isinstance(actual, np.ndarray):
            actual = actual.astype(float)
            actual[np.isnan(actual)] = 0.0
            
        elif isinstance(predicted, np.ndarray):
            predicted = predicted.astype(float)
            predicted[np.isnan(predicted)] = 0.0
            
        elif isinstance(actual, pd.core.series.Series):
            actual = actual.values 
            actual = actual.astype(float)
            actual[np.isnan(actual)] = 0.0
            
        elif isinstance(predicted, pd.core.series.Series):
            predicted = predicted.values
            predicted = predicted.astype(float)
            predicted[np.isnan(predicted)] = 0.0
            
        #separate logic if input is dataframe
        elif isinstance(actual, pd.core.frame.DataFrame):
            actual = actual.apply(pd.to_numeric).fillna(0)
            predicted = predicted.apply(pd.to_numeric).fillna(0)
            #calculate accuracy metrics for all series if set to true
            if separate_series:
                output_series = []
                for i in actual:
                    forecasted_dict = dask.delayed(accuracy)(actual[i],predicted[i])
                    output_series.append(forecasted_dict)
                total = dask.delayed(output_series).compute()
                forecast_df = pd.DataFrame(total,index=actual.columns).T
            else:
                actual = actual.values.flatten()
                predicted = predicted.values.flatten()
                
        else:
            raise TypeError('your inputs must be formatted as a numpy ndarray, pandas series or dataframe')
    except TypeError:
        print('your inputs must be formatted as a numpy ndarray, pandas series or dataframe')

    
    MAPE = mean_absolute_percentage_error(actual,predicted)
    SMAPE = s_mean_absolute_percentage_error(actual,predicted)
    ME = mean_error(actual,predicted)
    MAE = mean_absolute_error(actual,predicted)
    MSE = mean_squared_error(actual,predicted)
    RMSE = root_mean_squared_error(actual,predicted)
    SSE = sum_of_squared_error(actual,predicted)
    ThielsU = theil_u_statistic(actual,predicted)
    ACF1 = autocorrelation_lag_1(actual,predicted)
        
    accuracy_dict = {'MAPE':MAPE,'SMAPE':SMAPE,'ME':ME,'MAE':MAE,'MSE':MSE,'RMSE':RMSE,'ThielsU':ThielsU,'ACF1':ACF1}
    if separate_series:
        #return dataframe instead of separate series is true
        return forecast_df
    return accuracy_dict


# ----------------------------Scale independent metrics------------------------

# The mean absolute percentage error (MAPE), is a measure of prediction
# accuracy of a forecasting method in statistics. The 'min_val' variable is
# used to fill zeros in test data, if any. Filling zero with a constant helps
# to avoid division by zero error
def mean_absolute_percentage_error(y_true, y_pred, min_val=1):
    y_true = y_true + min_val
    return np.mean(np.abs((y_true - y_pred) / y_true)) * 100


# SMAPE is an alternative for MAPE when there are zeros in the testing data. It
# scales the absolute percentage by the sum of forecast and observed values
def s_mean_absolute_percentage_error(y_true, y_pred):
    return np.mean(np.abs((y_true - y_pred) / ((y_true + y_pred)/2))) * 100

#returns average error
def mean_error(y_true,y_pred):
    return np.mean(y_true - y_pred)

# MAE is used to measure how close the forecast is to observed value
def mean_absolute_error(y_true, y_pred):
    return np.mean(abs(y_true - y_pred))

# -----------------------------Scale dependent metrics-------------------------


# In statistics, the mean squared error (MSE) of an estimator measures the
# average of the squares of the errors. The squaring is necessary to remove any
# negative signs. It also gives more weight to larger differences
def mean_squared_error(y_true, y_pred):
    return np.mean((y_true - y_pred)**2)


# RMSE is the standard deviation of the residuals. It shares the same
# properties as MSE. The square root is used to dampen the magnitude of errors
# caused by squaring them.
def root_mean_squared_error(y_true, y_pred):
    return np.sqrt(np.mean((y_true - y_pred)**2))


# It measures the total squared deviation of forecasted observations, from the
# observed values
def sum_of_squared_error(y_true, y_pred):
    return np.sum((y_true - y_pred)**2)


# Thiels U accuracy measure (lies between 0  and 1, closer to 0 is more accurate)
def theil_u_statistic(y_true, y_pred):
    return np.sqrt(np.sum((y_pred - y_true)**2)/np.sum(y_true**2))

def autocorrelation_lag_1(y_true, y_pred, lag=1):
    error = y_true - y_pred
    # Slice the relevant subseries based on the lag
    y1 = error[:(len(error)-lag)]
    y2 = error[lag:]
    # Subtract the mean of the whole series x to calculate Cov
    sum_product = np.sum((y1-np.mean(error))*(y2-np.mean(error)))
    # Normalize with var of whole series
    return sum_product / ((len(error) - lag) * np.var(error))



import os
import sys
sys.path.insert(0, os.path.abspath('.'))
import sphinx_doc

from setuptools import setup

setup(name='magi',
      version=sphinx_doc.__version__,
      description='parallelized univariate time series forecasting library for python',
      long_description='This package is provides a python wrapper around other time series analysis libraries such as the forecast package in R and the Prophet library. This new layer of abstraction makes it very simple to put many different types of univariate time series models into production by using Dask as the parallelization layer',
      url='http://github.com/DavisTownsend/forecast',
      author= sphinx_doc.__author__,
      author_email='dtownsend@ea.com',
      classifiers=[
            'Development Status :: 2 - Pre-Alpha',
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6'
      ],
      keywords='time series analysis forecast forecasting predict model parallel',
      license='MIT',
      packages=[sphinx_doc.__name__],
      python_requires='~=3.5',
      install_requires=['numpy','pandas','dask','distributed','pystan','rpy2','fbprophet','plotly','cufflinks'],
      setup_requires=['pytest-runner'],
      tests_require=['pytest'],
      zip_safe=False)
